/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2005 Alpha Data
**
** methods_dma.c - Model-specific routines for DMA transfers.
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

static void
dumpDescriptorsPlxbased(
  DeviceContext* pDevCtx,
  PLXDESC* pDesc,
  os_uint32* pDescPci)
{
  unsigned int i = 0;
  os_uint32 pci, local, count, link;  

  os_dbgprint(0, ("Dump of PLXDESCs:\n"));
#if defined(CONFIG_64BIT) && CONFIG_64BIT
  os_dbgprint(0, ("### DESC-VIRT--------- DESC-BUS---------- [PCI-------- LOCAL----- COUNT----- LINK------]\n"));
#else
  os_dbgprint(0, ("### DESC-VIRT- DESC-BUS---------- [PCI-------- LOCAL----- COUNT----- LINK------]\n"));
#endif

  while (1) {
    pci = os_le32_to_cpu(pDesc->pci);
    local = os_le32_to_cpu(pDesc->local);
    count = os_le32_to_cpu(pDesc->count);
    link = os_le32_to_cpu(pDesc->link);
 
    os_dbgprint(0, ("%3.3ld 0x%p 0x%8lx [%8.8lx %8.8lx %8.8lx %8.8lx]\n",
		    (unsigned long)i,
		    (void*)pDesc,
		    (unsigned long)pDescPci[i],
		    (unsigned long)pci,
		    (unsigned long)local,
		    (unsigned long)count,
		    (unsigned long)link));
    if (link & PLXDESCF_ENDCHAIN) {
      break;
    }
    pDesc++;
    i++;
  }
}

static void
dumpDescriptorsXplbased(
  DeviceContext* pDevCtx,
  XPLDMADESC* pDesc,
  os_uint32* pDescPci)
{
  unsigned int i = 0;
  os_uint32 pah, pal, la, siz, ndh, ndl;
	
  os_dbgprint(0, ("Dump of DMA linked list:\n"));
#if defined(CONFIG_64BIT) && CONFIG_64BIT
  os_dbgprint(0, ("### DESC-VIRT--------- DESC-BUS---------- [PCI-------- LOCAL----- COUNT----- LINK------]\n"));
#else
  os_dbgprint(0, ("### DESC-VIRT- DESC-BUS---------- [PCI-------- LOCAL----- COUNT----- LINK------]\n"));
#endif
	
  while (1) {
    pal = os_le32_to_cpu(pDesc->PAL);
    pah = os_le32_to_cpu(pDesc->PAH);
    la = os_le32_to_cpu(pDesc->LA);
    siz = os_le32_to_cpu(pDesc->SIZ);
    ndl = os_le32_to_cpu(pDesc->NDL);
    ndh = os_le32_to_cpu(pDesc->NDH);
    os_dbgprint(0, ("%3.3ld 0x%p 0x%8.8lx { 0x%8.8lx%8.8lx 0x%8.8lx 0x%8.8lx 0x%8.8lx%8.8lx }\n",
		    (unsigned long)i,
		    (void*)pDesc,
		    (unsigned long)pDescPci[i],
		    (unsigned long)pah, 
		    (unsigned long)pal,
		    (unsigned long)la,
		    (unsigned long)siz,
		    (unsigned long)ndh,
		    (unsigned long)ndl));
    if (ndl & XPLDMADESC_ENDCHAIN) {
      break;
    }
    pDesc++;
    i++;
  }
}

/*
** -----------------------------------------------------------------------
** MethodAddDmaBuffer
** -----------------------------------------------------------------------
*/

MethodResult
methodAddDmaBufferPlxbased(
  DeviceContext* pDevCtx,
  unsigned int channel,
  unsigned int index,
  os_boolean bDirToLocalBus,
  DmaMapping* pMapping)
{
  DmaControl* pControl;
  PLXDESC* pDesc;
  os_uint32 flags;
  unsigned int which;

  /* Set up PLX descriptor */
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;
  pDesc = &pControl->descriptors[which].aligned.plxbased[index];
  if (index) {
    (pDesc - 1)->link |= os_cpu_to_le32(pControl->descriptors[which].pci[index]);
  }
  pDesc->pci = os_cpu_to_le32(pMapping->logical);
  pDesc->local = os_cpu_to_le32(pMapping->local);
  pDesc->count = os_cpu_to_le32(pMapping->length);
  flags = bDirToLocalBus ? 0x1U : 0x9U;
  pDesc->link = os_cpu_to_le32(flags);

  return MethodResult_Success;
}

MethodResult
methodAddDmaBufferXplrev0(
  DeviceContext* pDevCtx,
  unsigned int channel,
  unsigned int index,
  os_boolean bDirToLocalBus,
  DmaMapping* pMapping)
{
  return MethodResult_NotSupported;
}

MethodResult methodAddDmaBufferXplrev1(
  DeviceContext* pDevCtx,
  unsigned int channel,
  unsigned int index,
  os_boolean bDirToLocalBus,
  DmaMapping* pMapping)
{
  DmaControl* pControl;
  XPLDMADESC* pDesc;
  os_uint32 sizFlags;
  unsigned int which;

  /* Set up PLX descriptor */
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;
  pDesc = &pControl->descriptors[which].aligned.xpl[index];
  if (index) {
    (pDesc - 1)->NDL |= os_cpu_to_le32(pControl->descriptors[which].pci[index]);
  }
  if (bDirToLocalBus) {
    sizFlags = XPLDMADESC_PCITOLOCAL;
  } else {
    sizFlags = XPLDMADESC_LOCALTOPCI;
  }
  pDesc->PAL = os_cpu_to_le32((os_uint32)pMapping->logical);
  pDesc->PAH = os_cpu_to_le32(0);
  pDesc->LA = os_cpu_to_le32(pMapping->local);
  pDesc->SIZ = os_cpu_to_le32(pMapping->length | sizFlags);
  pDesc->NDL = os_cpu_to_le32(XPLDMADESC_PCISPACE);
  pDesc->NDH = os_cpu_to_le32(0);
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodDoDma
** -----------------------------------------------------------------------
*/

static MethodResult
methodDoDmaPlxbased(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  os_uint16 shadow_dmacsr,
  unsigned int numBuffer)
{
  PLX9080REGS* pRegs = pDevCtx->hw.adapter.pci9080;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  DmaControl* pControl;
  unsigned int which;
  MethodResult res = MethodResult_Success;
  os_uint16 mask, val16;
  PLXDESC* pDesc;

  os_dbgprint(6, ("methodDoDmaPlxbased: channel=%ld bStart=%ld mode=0x%lx, shadow.dmacsr=0x%4.4lx\n",
		  (unsigned long)channel,
		  (unsigned long)bStart,
		  (unsigned long)mode,
		  (unsigned long)shadow_dmacsr));

  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;

  if (bStart) {
    /*
    ** Kick off a chunk of a DMA transfer
    */

    pDesc = &pControl->descriptors[which].aligned.plxbased[numBuffer - 1];
    pDesc->link |= os_cpu_to_le32(PLXDESCF_ENDCHAIN | PLXDESCF_INTERRUPT);

    if (os_dbglevel(8)) {
      dumpDescriptorsPlxbased(pDevCtx,
			      pControl->descriptors[which].aligned.plxbased,
			      pControl->descriptors[which].pci);
    }
  
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    mask = channel ? 0x0200U : 0x0002U;
    val16 = os_read_register_le16(handle, &pRegs->DMACSR.asWord);
    if (val16 & (mask << 3)) {
      if (numBuffer > 1) {
        /* Set current transfer mode (chained mode bit added) */
        os_write_register_le32(handle, &pRegs->DMA[channel].MODE, mode | (0x1U << 9));
        /* Set current transfer size to 0 */
        os_write_register_le32(handle, &pRegs->DMA[channel].SIZE, 0);
        /* Set the descriptor pointer */
        os_write_register_le32(handle, &pRegs->DMA[channel].DPR,
                               pControl->descriptors[which].pci[0] | 0x1U);
      } else {
        pDesc = &pControl->descriptors[which].aligned.plxbased[0];
        /* Set current transfer mode (chained mode bit NOT added) */
        os_write_register_le32(handle, &pRegs->DMA[channel].MODE, mode);
        /* Set current transfer PCI address */
        os_write_register_le32(handle, &pRegs->DMA[channel].PADR, os_le32_to_cpu(pDesc->pci));
        /* Set current transfer local address */
        os_write_register_le32(handle, &pRegs->DMA[channel].LADR, os_le32_to_cpu(pDesc->local));
        /* Set current transfer size */
        os_write_register_le32(handle, &pRegs->DMA[channel].SIZE, os_le32_to_cpu(pDesc->count));
        /* Set next descriptor pointer */
        os_write_register_le32(handle, &pRegs->DMA[channel].DPR, os_le32_to_cpu(pDesc->link));
      }
      /* Make sure the writes have completed */
      os_read_register_le32(handle, &pRegs->DMA[channel].DPR);
      /* Strobe the start bit */
      os_write_register_le16(handle, &pRegs->DMACSR.asWord, shadow_dmacsr | mask);
    } else {
      os_dbgprint(0, ("*** methodDoDmaPlxbased: DMA channel %ld not idle\n", (unsigned long) channel));
      res = MethodResult_GeneralFailure;
    }
  } else {
    /*
    ** Abort a chunk of a DMA transfer
    */

    mask = channel ? 0x0100 : 0x0001;
  
    /* First, disable the DMA channel */
    os_write_register_le16(handle, &pRegs->DMACSR.asWord, shadow_dmacsr & ~mask);
    
    /* Now check if DMA channel is active or not */
    val16 = os_read_register_le16(handle, &pRegs->DMACSR.asWord);
    if (val16 & (mask << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaPlxbased: nothing to abort on DMA channel %ld\n", (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le16(handle, &pRegs->DMACSR.asWord, shadow_dmacsr | (mask << 2));
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(50);
      /* Did we succeed in aborting it? */
      val16 = os_read_register_le16(handle, &pRegs->DMACSR.asWord);
      if (val16 & (mask << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(3, ("methodDoDmaPlxbased: aborted DMA on channel %ld\n", (unsigned long) channel));
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaPlxbased: failed to abort DMA on channel %ld!\n", (unsigned long) channel));
	res = MethodResult_GeneralFailure;
      }
    }
    
    /*
    ** Reenable the DMA channel and clear the DMA interrupt that will have
    ** occurred if we aborted.
    */
    os_write_register_le16(handle, &pRegs->DMACSR.asWord, shadow_dmacsr | mask | (mask << 3));
  }

  return res;
}

MethodResult
methodDoDmaXrc(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  unsigned int numBuffer)
{
  return methodDoDmaPlxbased(pDevCtx, channel, bStart, mode, pDevCtx->shadow.regs.plxbased.dmacsr, numBuffer);
}

MethodResult
methodDoDmaXplrev0(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  unsigned int numBuffer)
{
  return MethodResult_NotSupported;
}

MethodResult
methodDoDmaXplrev1(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  unsigned int numBuffer)
{
  XPLBRIDGEREGS* pRegs = pDevCtx->hw.adapter.xpl;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  DmaControl* pControl;
  unsigned int which;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* pDesc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;

  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;

  if (bStart) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    pDesc = &pControl->descriptors[which].aligned.xpl[numBuffer - 1];
    pDesc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pControl->descriptors[which].aligned.xpl,
			      pControl->descriptors[which].pci);
    }
    
    descBus = pControl->descriptors[which].pci[0];
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
        /* Set current transfer mode (adds in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode | (0x1U << 9));
        /* Set the descriptor pointer & flags */
        os_write_register_le32(handle, &pRegs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
        os_write_register_le32(handle, &pRegs->DMA[channel].NDH, 0);
      } else {
        pDesc = &pControl->descriptors[which].aligned.xpl[0];
        /* Set current transfer mode (DOESN'T add in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode);
        /* Set current transfer PCI address */
        os_write_register_le32(handle, &pRegs->DMA[channel].PAL, os_le32_to_cpu(pDesc->PAL));
        os_write_register_le32(handle, &pRegs->DMA[channel].PAH, os_le32_to_cpu(pDesc->PAH));
        /* Set current transfer local address */
        os_write_register_le32(handle, &pRegs->DMA[channel].LA, os_le32_to_cpu(pDesc->LA));
        /* Set current transfer size */
        os_write_register_le32(handle, &pRegs->DMA[channel].SIZ, os_le32_to_cpu(pDesc->SIZ));
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaXplrev1: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaXplrev1: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaXplrev1: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaXplrev1: aborted, channel=%ld\n",
			(unsigned long) channel));
	/* Clear the DMA interrupt that will have occurred */
	val32 = 0x100U << channel;
	os_write_register_le32(handle, &pRegs->PISTAT, val32);
	os_read_register_le32(handle, &pRegs->PISTAT);
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaXplrev1: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /* Reenable the DMA channel. */
    dxctl |= 0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}

MethodResult
methodDoDmaXrc4ls(
  DeviceContext*  pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  unsigned int numBuffer)
{
  return methodDoDmaPlxbased(pDevCtx, channel, bStart, mode, pDevCtx->shadow.regs.xrc4ls.dmacsr, numBuffer);
}

MethodResult
methodDoDmaXrce4fx(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  unsigned int numBuffer)
{
  XPLBRIDGEREGS* pRegs = pDevCtx->hw.adapter.xpl;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  DmaControl* pControl;
  unsigned int which;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* pDesc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;
  
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;

  if (bStart) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    pDesc = &pControl->descriptors[which].aligned.xpl[numBuffer - 1];
    pDesc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pControl->descriptors[which].aligned.xpl,
			      pControl->descriptors[which].pci);
    }
    
    descBus = pControl->descriptors[which].pci[0];
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
        /* Set current transfer mode (adds in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode | (0x1U << 9));
        /* Set the descriptor pointer & flags */
        os_write_register_le32(handle, &pRegs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
        os_write_register_le32(handle, &pRegs->DMA[channel].NDH, 0);
      } else {
        pDesc = &pControl->descriptors[which].aligned.xpl[0];
        /* Set current transfer mode (DOESN'T add in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode);
        /* Set current transfer PCI address */
        os_write_register_le32(handle, &pRegs->DMA[channel].PAL, os_le32_to_cpu(pDesc->PAL));
        os_write_register_le32(handle, &pRegs->DMA[channel].PAH, os_le32_to_cpu(pDesc->PAH));
        /* Set current transfer local address */
        os_write_register_le32(handle, &pRegs->DMA[channel].LA, os_le32_to_cpu(pDesc->LA));
        /* Set current transfer size */
        os_write_register_le32(handle, &pRegs->DMA[channel].SIZ, os_le32_to_cpu(pDesc->SIZ));
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.regs.xrce4fx.dxctl[channel];
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaXplrev1: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaXplrev1: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.regs.xrce4fx.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaXplrev1: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaXplrev1: aborted, channel=%ld\n",
			(unsigned long) channel));
	/* Clear the DMA interrupt that will have occurred */
	val32 = 0x100U << channel;
	os_write_register_le32(handle, &pRegs->PISTAT, val32);
	os_read_register_le32(handle, &pRegs->PISTAT);
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaXplrev1: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /* Reenable the DMA channel. */
    dxctl |= 0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}

MethodResult
methodDoDmaXrc4fx(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  unsigned int numBuffer)
{
  XRC4FXBRIDGEREGS* pRegs = pDevCtx->hw.adapter.xrc4fx;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  DmaControl* pControl;
  unsigned int which;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* pDesc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;
  
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;

  if (bStart) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    pDesc = &pControl->descriptors[which].aligned.xpl[numBuffer - 1];
    pDesc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pControl->descriptors[which].aligned.xpl,
			      pControl->descriptors[which].pci);
    }
    
    descBus = pControl->descriptors[which].pci[0];
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
        /* Set current transfer mode (adds in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode | (0x1U << 9));
        /* Set the descriptor pointer & flags */
        os_write_register_le32(handle, &pRegs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
        os_write_register_le32(handle, &pRegs->DMA[channel].NDH, 0);
      } else {
        pDesc = &pControl->descriptors[which].aligned.xpl[0];
        /* Set current transfer mode (DOESN'T add in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode);
        /* Set current transfer PCI address */
        os_write_register_le32(handle, &pRegs->DMA[channel].PAL, os_le32_to_cpu(pDesc->PAL));
        os_write_register_le32(handle, &pRegs->DMA[channel].PAH, os_le32_to_cpu(pDesc->PAH));
        /* Set current transfer local address */
        os_write_register_le32(handle, &pRegs->DMA[channel].LA, os_le32_to_cpu(pDesc->LA));
        /* Set current transfer size */
        os_write_register_le32(handle, &pRegs->DMA[channel].SIZ, os_le32_to_cpu(pDesc->SIZ));
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.regs.xrc4fx.dxctl[channel];
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaXrc4fx: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaXrc4fx: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.regs.xrc4fx.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaXrc4fx: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaXrc4fx: aborted, channel=%ld\n",
			(unsigned long) channel));
	/* Clear the DMA interrupt that will have occurred */
	val32 = 0x100U << channel;
	os_write_register_le32(handle, &pRegs->PISTAT, val32);
	os_read_register_le32(handle, &pRegs->PISTAT);
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaXrc4fx: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /* Reenable the DMA channel. */
    dxctl |= 0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}

MethodResult
methodDoDmaBbp(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean bStart,
  os_uint32 mode,
  unsigned int numBuffer)
{
  XRC4FXBRIDGEREGS* pRegs = pDevCtx->hw.adapter.xrc4fx;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  DmaControl* pControl;
  unsigned int which;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* pDesc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;
  
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;

  if (bStart) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    pDesc = &pControl->descriptors[which].aligned.xpl[numBuffer - 1];
    pDesc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pControl->descriptors[which].aligned.xpl,
			      pControl->descriptors[which].pci);
    }
    
    descBus = pControl->descriptors[which].pci[0];
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
        /* Set current transfer mode (adds in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode | (0x1U << 9));
        /* Set the descriptor pointer & flags */
        os_write_register_le32(handle, &pRegs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
        os_write_register_le32(handle, &pRegs->DMA[channel].NDH, 0);
      } else {
        pDesc = &pControl->descriptors[which].aligned.xpl[0];
        /* Set current transfer mode (DOESN'T add in chained-mode bit) */
        os_write_register_le32(handle, &pRegs->DMA[channel].CFG, mode);
        /* Set current transfer PCI address */
        os_write_register_le32(handle, &pRegs->DMA[channel].PAL, os_le32_to_cpu(pDesc->PAL));
        os_write_register_le32(handle, &pRegs->DMA[channel].PAH, os_le32_to_cpu(pDesc->PAH));
        /* Set current transfer local address */
        os_write_register_le32(handle, &pRegs->DMA[channel].LA, os_le32_to_cpu(pDesc->LA));
        /* Set current transfer size */
        os_write_register_le32(handle, &pRegs->DMA[channel].SIZ, os_le32_to_cpu(pDesc->SIZ));
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.regs.bbp.dxctl[channel];
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaBbp: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaBbp: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.regs.bbp.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaBbp: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(handle, &pRegs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaBbp: aborted, channel=%ld\n",
			(unsigned long) channel));
	/* Clear the DMA interrupt that will have occurred */
	val32 = 0x100U << channel;
	os_write_register_le32(handle, &pRegs->PISTAT, val32);
	os_read_register_le32(handle, &pRegs->PISTAT);
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaBbp: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /* Reenable the DMA channel. */
    dxctl |= 0x4U;
    os_write_register_le32(handle, &pRegs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}
